/*
 * ALSA Patch Bay
 *
 * Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *
 * You have permission to use this file under the GNU General
 * Public License, version 2 or later.  See the file COPYING
 * for the full text.
 *
 */

#include "apb.h"
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "driver/apb-alsa.h"
#ifdef HAVE_GTKMM
#include "gui/apb-gtkmm.h"
#endif
#ifdef HAVE_FLTK
#include "gui/apb-fltk.h"
#endif

APB::DebugLevel debugLevel = APB::DEBUG;

namespace APB {

static void * xmalloc (size_t size) {
  void * ptr;
  ptr = malloc (size);
  if (!ptr) {
    exit (1);
  }
  return ptr;
}

static void * xrealloc (void * mem, size_t size) {
  void * ptr;
  ptr = realloc (mem, size);
  if (!ptr) {
    exit (1);
  }
  return ptr;
}


std::string Misc::int2string (int i) {
  static uint size = 16;
  static char * s = NULL;
  int err = -1;

  if (!s) s = (char *) xmalloc (size * sizeof(char));

  while (err == -1) {
    err = snprintf (s, size, "%d", i);
    if (err == -1) {
      size += 16;
      s = (char *) xrealloc (s, size);
    }
  }

  return std::string (s);
}

int Misc::randomNumber (double lower, double upper) {
  return (int) ((lower-1)+(int) ((upper+1)*rand()/(RAND_MAX+1.0)));
}

/////////////////////////////////////////
//////////// Subscription ///////////////
/////////////////////////////////////////

bool Subscription::operator == (const Subscription& sub) const {
  if (!(sub.to()->equals (_to) && sub.from()->equals (_from))) {
    return false;
  }

  std::map<std::string, int>::const_iterator that_it;
  for (std::map<std::string, int>::const_iterator this_it = _attributes.begin(),
       that_it = sub._attributes.begin();
       this_it != _attributes.end();
       ++this_it, ++that_it) {
    if (this_it->first != that_it->first ||
        this_it->second != this_it->second) {
      return false;
    }
  }
  
  if (that_it != sub._attributes.end()) {
    return false;
  }
}

/////////////////////////////////////////
//////////// AlsaPatchBay ///////////////
/////////////////////////////////////////


Driver * AlsaPatchBay::newDriver (std::string title, int argc, char ** argv, DriverType type) {
  switch (type) {
    case ALSA:
      return new Alsa::Driver (title, argc, argv);
      break;
  }
}

UI * AlsaPatchBay::newUI (std::string title, int argc, char ** argv, UIType type, Driver * driver) {
  switch (type) {
#ifdef HAVE_GTKMM
    case UI_GTKMM:
      return new GTKmm::UI (title, argc, argv, driver);
      break;
#endif
#ifdef HAVE_FLTK
    case UI_FLTK:
      return new FLTK::UI (title, argc, argv, driver);
      break;
#endif
  }
}

} /* namespace APB */

void print_help () {
  printf("%s version %s\n", PACKAGE_NAME, PACKAGE_VERSION);
  printf("Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)\n");
  printf("\n");
  printf("This program comes with ABSOLUTELY NO WARRANTY.  You are licensed to use it\n");
  printf("under the terms of the GNU General Public License, version 2 or later.  See\n");
  printf("the COPYING file that came with this software for details.\n");
  printf("\n");
  printf(" -h, --help         Display this help info\n");
  printf(" -u, --ui <UI>      Specify which user interface to use.\n");
  printf("                    <UI> can be one of: ");
#ifdef HAVE_GTKMM
  printf("GTKmm ");
#endif
#ifdef HAVE_FLTK
  printf("FLTK ");
#endif
  printf("\n\n");
}

int main(int argc, char ** argv) {
  srand (time(0));
  using namespace APB;
  AlsaPatchBay * apb;
  const char * options = "hu:";
  struct option long_options[] = {
    { "help", 0, NULL, 'h' },
    { "ui", 1, NULL, 'u' },
    { 0, 0, 0, 0 }
  };
#ifdef HAVE_GTKMM
  UIType uitype = UI_GTKMM;
#else
  UIType uitype = UI_FLTK;
#endif
  int opt;

  while ((opt = getopt_long (argc, argv, options, long_options, NULL)) != -1)
    {
      switch (opt)
        {
          case 'h':
            print_help ();
            exit (0);
            break;
          case 'u':
#ifdef HAVE_GTKMM
            if (strcasecmp (optarg, "GTKmm") == 0)
              {
                uitype = UI_GTKMM;
                break;
              }
#endif
#ifdef HAVE_FLTK
            if (strcasecmp (optarg, "FLTK") == 0)
              {
                uitype = UI_FLTK;
                break;
              }
#endif
            printf ("unknown ui '%s'\n", optarg);
            print_help ();
            exit (1);
            break;
          case ':':
          case '?':
            print_help ();
            exit (1);
            break;
        }
    }
  
  try {
    apb = new AlsaPatchBay ("ALSA Patch Bay", argc, argv, ALSA, uitype);
  } catch (Exception e) {
    std::cerr << std::endl << "Caught exception while creating AlsaPatchBay: " << e.desc << std::endl;
    return 1;
  }
  apb->run ();
  return 0;
}

