/*
 * ALSA Patch Bay
 *
 * Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *
 * You have permission to use this file under the GNU General
 * Public License, version 2 or later.  See the file COPYING
 * for the full text.
 *
 */

#ifndef __APB_APB_H__
#define __APB_APB_H__

#include <list>
#include <string>
#include <vector>
#include <map>

#include "config.h"

#define DEBUG_STRING std::string(__FILE__) + ":" + Misc::int2string(__LINE__) + ":" + \
      __FUNCTION__ + ": "
      
      
namespace APB {

typedef enum { ERROR = 0,
               WARNING,
               NOTICE,
               DEBUG } DebugLevel;

struct Exception {
  std::string desc;
  int err;
  Exception (std::string d) : desc (d) {}
  Exception (std::string d, int e) : desc (d), err (e) {}
};

class Misc {
  public:
    static std::string int2string (int);
    static int randomNumber (double, double);
};


/***********************************************/
/************** Driver interface ***************/
/***********************************************/

typedef enum { ALSA, JACK } DriverType;

class Driver;

class Addr {
  public:
    virtual bool equals (const Addr *) const = 0;
    virtual Driver * driver () const = 0;
    virtual std::string getName () const = 0;
};

class Subscription {
  private:
    const Addr * _from;
    const Addr * _to;
    std::map<std::string, int> _attributes;
  public:
    Subscription (const Addr * from, const Addr * to)
      : _from (from), _to (to) {}
    Subscription (const Subscription& subscription)
      : _from (subscription._from),
        _to (subscription._to) {}
    const Addr * from () const { return _from; }
    const Addr * to () const { return _to; }
    std::string getName () {
      return std::string("'") + _from->getName() + "' -> '" + _to->getName() + "'";
    }
    bool operator == (const Subscription& sub) const;
    const std::map<std::string, int>& getAttributes () const { return _attributes; }
    void setAttribute (std::string str, int num) { _attributes[str] = num; }
};

class UI;

class Driver {
  public:
    virtual std::string findClientName (const Addr *) const = 0;
    virtual std::string findPortName (const Addr *) const = 0;
    virtual const std::list<Addr *>& getReadPorts () = 0;
    virtual const std::list<Addr *>& getWritePorts () = 0;
    virtual const std::list<Subscription>& getSubscriptions () = 0;
    virtual void refreshPorts () = 0;
    virtual void refreshSubscriptions () = 0;
    virtual void subscribePorts (const Addr *, const Addr *) = 0;
    virtual void unsubscribePorts (const Addr *, const Addr *) = 0;
    virtual void setJFD (bool) = 0;
    virtual void setUI (UI *) = 0;
};


/*******************************************/
/************** UI interface ***************/
/*******************************************/

typedef enum {
#ifdef HAVE_GTKMM
  UI_GTKMM,
#endif
#ifdef HAVE_FLTK
  UI_FLTK
#endif
} UIType;

#ifdef HAVE_GTKMM
# define DEFAULT_UI UI_GTKMM
#else
# define DEFAULT_UI UI_FLTK
#endif

typedef enum { ONLYONE, ANY } ChoiceType;

struct Choice {
  std::string _choice;
  bool _takesInt;
  int _value;
  Choice (//int id,
          const std::string& choice,
          bool takesInt = false,
          int defaultValue = 0)
    : _choice (choice),
      _takesInt (takesInt),
      _value (defaultValue) {}
  Choice (const Choice& choice)
    : _choice (choice._choice),
      _takesInt (choice._takesInt),
      _value (choice._value) {}
};

struct ChoiceSpec {
  std::string _message;
  ChoiceType _type;
  std::list<Choice> _choices;
  ChoiceSpec (std::string message, ChoiceType type)
    : _message (message), _type (type) {}
  void addChoice (const Choice& choice) {
    _choices.push_back (choice);
  }
};

typedef std::pair< bool, int > ChoiceResult; // (selected, value)
                                              	    	  	

class UI {
  public:
    virtual void log (DebugLevel, const std::string&) = 0;
    virtual void run () = 0;
    virtual std::vector<ChoiceResult> getChoice (std::string, const ChoiceSpec&) = 0;
};

/*******************************************/
/************** AlsaPatchBay ***************/
/*******************************************/

class AlsaPatchBay {
  private:
    UI * _ui;
    Driver * _driver;
    static Driver * newDriver (std::string, int, char **, DriverType);
    static UI * newUI (std::string, int, char **, UIType, Driver *);
  public:
    AlsaPatchBay (std::string title, int argc, char ** argv,
                  DriverType driverType = ALSA,
                  UIType uiType = DEFAULT_UI) {
      _driver = newDriver (title, argc, argv, driverType);
      _ui = newUI (title, argc, argv, uiType, _driver);
      _driver->setUI (_ui);
    }
    void run () {
      _ui->run ();
    }
};

} /* namespace APB */

extern APB::DebugLevel debugLevel;

#endif /* __APB_APB_H__ */
