/*
 * ALSA Patch Bay
 *
 * Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *
 * You have permission to use this file under the GNU General
 * Public License, version 2 or later.  See the file COPYING
 * for the full text.
 *
 */

#include "apb-alsa.h"

namespace APB {
namespace Alsa {


/////////////////////////////////
/////////// Addr ////////////////
/////////////////////////////////

std::string Addr::getName () const {
  const static short break_size = 50;
  std::string s(_driver->findClientName (this) + "::" + _driver->findPortName (this));
  if (s.size() < break_size) {
    return s;
  } else {
    return _driver->findClientName (this) + "::\n" + _driver->findPortName (this);
  }
}


///////////////////////////////////
/////////// Driver ////////////////
///////////////////////////////////

Driver::Driver (std::string title, int argc, char ** argv)
  : _jfd (true),
    _ui (0) {
  int err;
  err = snd_seq_open(&_seq, "default",
                     SND_SEQ_OPEN_DUPLEX,
                     SND_SEQ_NONBLOCK);
  if (err) {
    std::string s (DEBUG_STRING + "couldn't open alsa sequencer");
    throw APB::Exception (s);
  }
}

std::string Driver::findClientName (const APB::Addr * addr) const {
  snd_seq_client_info_t * cinfo;
  snd_seq_client_info_alloca (&cinfo);
  int err = snd_seq_get_any_client_info (((Driver *)addr->driver())->_seq, ((const Addr *)addr)->client(), cinfo);
  if (err) {
    throw APB::Exception (std::string("could not find client in order to resolve its name:\n ")
                          + snd_strerror (err));
  }
  return std::string (snd_seq_client_info_get_name (cinfo));
}

std::string Driver::findPortName (const APB::Addr * addr) const {
  snd_seq_port_info_t * pinfo;
  snd_seq_port_info_alloca (&pinfo);
  int err = snd_seq_get_any_port_info (((Driver *)addr->driver())->_seq,
            ((const Addr *)addr)->client(), ((const Addr *)addr)->port(), pinfo);
  if (err) {
    throw APB::Exception (std::string("could not find port in order to resolve its name:\n ")
                          + snd_strerror (err));
  }
  return std::string (snd_seq_port_info_get_name (pinfo));
}

void Driver::refreshPorts () {
  refreshPorts (_readPorts, SND_SEQ_PORT_CAP_READ | SND_SEQ_PORT_CAP_DUPLEX);
  refreshPorts (_writePorts, SND_SEQ_PORT_CAP_WRITE | SND_SEQ_PORT_CAP_DUPLEX);
}

void Driver::refreshPorts (std::list<APB::Addr *>& portList, unsigned int capabilities) {
  portList.clear ();

  snd_seq_client_info_t * cinfo;
  snd_seq_client_info_alloca (&cinfo);
  snd_seq_client_info_set_client (cinfo, -1);

  snd_seq_port_info_t * pinfo;
  snd_seq_port_info_alloca (&pinfo);
  
  
  while (snd_seq_query_next_client (_seq, cinfo) >= 0) {

    snd_seq_port_info_set_client(pinfo, snd_seq_client_info_get_client(cinfo));
    snd_seq_port_info_set_port (pinfo, -1);

    while (snd_seq_query_next_port (_seq, pinfo) >= 0) {

      int caps = snd_seq_port_info_get_capability (pinfo);

      if (caps & capabilities) {
        Alsa::Addr * portAddr  = new Alsa::Addr (*snd_seq_port_info_get_addr (pinfo), this);
        portList.push_back (portAddr);
      }
    }
  }

}

const std::list<APB::Addr *>& Driver::getReadPorts () {
  return _readPorts;
}

const std::list<APB::Addr *>& Driver::getWritePorts () {
  return _writePorts;
}

const std::list<APB::Subscription>& Driver::getSubscriptions () {
  return _subscriptions;
}

void Driver::refreshSubscriptions () {
  for (std::list<APB::Subscription>::iterator s = _subscriptions.begin ();
       s != _subscriptions.end ();
       ++s) {
// FIXME: what to do with this dodgeyness?
//    free ((Addr *)s->to());
  }
  _subscriptions.clear ();

  snd_seq_query_subscribe_t * subsinfo;
  snd_seq_query_subscribe_alloca (&subsinfo);

  for (std::list<APB::Addr *>::iterator i = _readPorts.begin ();
       i != _readPorts.end ();
       ++i) {
    snd_seq_query_subscribe_set_root(subsinfo, &((Addr *)*i)->addr());
    snd_seq_query_subscribe_set_index (subsinfo, 0);

    while (!snd_seq_query_port_subscribers (_seq, subsinfo)) {
      snd_seq_addr_t addr = *snd_seq_query_subscribe_get_addr (subsinfo);
      APB::Addr * waddrp = 0;
      Addr waddr (addr, this);

      for (std::list<APB::Addr *>::iterator it = _writePorts.begin ();
           it != _writePorts.end ();
           ++it) {
        if ((*it)->equals (&waddr)) {
          waddrp = *it;
        }
      }

      Subscription sub (*i, waddrp);
      sub.setAttribute ("exclusive", snd_seq_query_subscribe_get_exclusive (subsinfo));
      sub.setAttribute ("time update", snd_seq_query_subscribe_get_time_update (subsinfo));
      sub.setAttribute ("realtime", snd_seq_query_subscribe_get_time_real (subsinfo));
      _subscriptions.push_back (sub);
      
      snd_seq_query_subscribe_set_index (subsinfo,
        snd_seq_query_subscribe_get_index (subsinfo) + 1);
    }
  }
}

void Driver::doPortSubscription (snd_seq_port_subscribe_t * subs,
                                 const APB::Addr * from,
                                 const APB::Addr * to) {
  if (snd_seq_get_port_subscription(_seq, subs) == 0) {
    // already subscribed
    _ui->log (WARNING, std::string("Subscription ") + from->getName() + " -> "
                        + to->getName() + " already exists");
    return;
  }
  int err = snd_seq_subscribe_port(_seq, subs);
  if (err < 0) {
    std::string s (std::string("Subscription ") + from->getName() + " -> "
              + to->getName() + " failed:\n  " + snd_strerror (err));
    throw Exception (s);
    return;
  }

  _subscriptions.push_back (Subscription (from, to));
  _ui->log (NOTICE, std::string("Subscription ") + from->getName() + " -> "
                        + to->getName() + " succeeded");
}

void Driver::doPortUnsubscription (snd_seq_port_subscribe_t * subs,
                                   const APB::Addr * from,
                                   const APB::Addr * to) {
  int err = snd_seq_get_port_subscription(_seq, subs);
  if (err != 0) {
    // already subscribed
    std::string s (std::string ("Subscription ") + Subscription (from, to).getName() + 
              " does not exist");
    throw Exception (s, err);
    return;
  }
  
  err = snd_seq_unsubscribe_port(_seq, subs);
  if (err < 0) {
    std::string s (std::string ("Could not remove subscription ") +
              Subscription (from, to).getName() + "\n  " +
              snd_strerror(err));
    throw Exception (s, err);
  }

  for (std::list<Subscription>::iterator it = _subscriptions.begin();
       it != _subscriptions.end();
       ++it) {
    if (it->to()->equals(to) && it->from()->equals(from)) {
// FIXME: memory leak
//      free ((Alsa::Addr*)it->to());
      break;
    }
  }
  _subscriptions.remove (Subscription(from, to));
  _ui->log (NOTICE, std::string("Unsubscription ") + from->getName() + " -> "
                        + to->getName() + " succeeded");
}

void Driver::subscribePorts (const APB::Addr * from,
                             const APB::Addr * to) {
  subscribePorts (from, to, true);
}

void Driver::unsubscribePorts (const APB::Addr * from,
                               const APB::Addr * to) {
  subscribePorts (from, to, false);
}

void Driver::subscribePorts (const APB::Addr * from,
                             const APB::Addr * to,
                             bool create) {

  snd_seq_port_subscribe_t * subs;
  snd_seq_port_subscribe_alloca(&subs);
  snd_seq_port_subscribe_set_sender(subs, &((Alsa::Addr *)from)->addr());
  snd_seq_port_subscribe_set_dest(subs, &((Alsa::Addr *)to)->addr());
  snd_seq_port_subscribe_set_queue(subs, 0);

  if (create) {
    ChoiceSpec choiceSpec ("Subscription options", ANY);
    std::vector<ChoiceResult> results;
    if (_jfd) {
      for (int j = 0; j < 3; ++j) {
        results.push_back (std::pair<bool, int> (false, 0));
      }
    } else {
      choiceSpec.addChoice (Choice ("Exclusive"));
      choiceSpec.addChoice (Choice ("Time update"));
      choiceSpec.addChoice (Choice ("Real time"));
      results = _ui->getChoice ("Subscription options", choiceSpec);
    }
    snd_seq_port_subscribe_set_exclusive(subs, (int) results[0].first);
    snd_seq_port_subscribe_set_time_update(subs, (int) results[1].first);
    snd_seq_port_subscribe_set_time_real(subs, (int) results[2].first);
    doPortSubscription (subs, from, to);
  } else {
    unsigned failCount = 0;
    int err = 0;
    for (int i = 0; i < 2; ++i) {
      for (int j = 0; j < 2; ++j) {
        for (int k = 0; k < 2; ++k) {
          snd_seq_port_subscribe_set_exclusive(subs, i);
          snd_seq_port_subscribe_set_time_update(subs, j);
          snd_seq_port_subscribe_set_time_real(subs, k);
          try {
            doPortUnsubscription (subs, from, to);
          }
          catch (Exception e) {
            err = e.err;
            failCount++;
          }
        }
      }
    }
    if (failCount == 8) {
      throw Exception (std::string("Unsubscribing '") +
                       from->getName() + "' -> '" + to->getName() +
                       "' failed.  Last error was:\n  " + snd_strerror (err));
    }
  }
}

void Driver::setJFD (bool jfd) {
  _jfd = jfd;
}

void Driver::setUI (APB::UI * ui) {
  _ui = ui;
}

} /* namespace Alsa */
} /* namespace APB */
