/*
 * ALSA Patch Bay
 *
 * Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *
 * You have permission to use this file under the GNU General
 * Public License, version 2 or later.  See the file COPYING
 * for the full text.
 *
 */

#ifndef __APB__GTKMM_H__
#define __APB__GTKMM_H__

#include <iostream>
#include <list>
#include <string>
#include <gtkmm-2.0/gtkmm.h>
#include <gtkmm/layout.h>
#include <gtkmm/stock.h>
#include <gtkmm/radiobutton.h>

#include "apb.h"

namespace APB {
namespace GTKmm {

using Glib::ustring;

class ChoiceWidget : public Gtk::HBox {
  private:
    static Gtk::RadioButton::Group _group;
    Gtk::ToggleButton * _toggleButton;
    Gtk::SpinButton * _spinButton;
    Gtk::Adjustment _spinButtonAdj;
    void setInput ();
  public:
    ChoiceWidget (ChoiceType, const Choice&);
    const ChoiceResult& getResult () const;
};

class ChoiceWindow : public Gtk::Dialog {
  private:
    std::list<ChoiceWidget *> _choiceWidgets;
  public:
    ChoiceWindow (const std::string&, const ChoiceSpec&);
    std::vector<ChoiceResult> getChoice ();
};
 
class ButtonArray;

class PortButton : public Gtk::Button {
  private:
    const Addr * _addr;
    ButtonArray * _buttonArray;
    bool _read;
    guint _index;
  protected:
    virtual void on_clicked ();
  public:
    PortButton (const Addr * addr,
                ButtonArray * buttonArray,
                bool read,
                guint index = 0)
      : Gtk::Button (addr->getName()),
        _addr (addr),
        _buttonArray (buttonArray),
        _read (read),
        _index (index) {}
    void addr (const Addr * addr) { _addr = addr; }
    const Addr * addr () { return _addr; }
    const Addr * addr () const { return _addr; }
    void index (guint index) { _index = index; }
    guint index () { return _index; }
    const guint index () const { return _index; }
};


class LogBox : public Gtk::TextView {
  private:
    Glib::RefPtr<Gtk::TextBuffer> _textBuffer;
  public:
    LogBox () {
     _textBuffer = Gtk::TextBuffer::create();
     set_buffer (_textBuffer);
    }
    void log (DebugLevel, const std::string&);
};

class Colour : public Gdk::Color {
  private:
    static Colour _currentColour;
    static const guint _incr = 10000;
  public:
    Colour () {}
    Colour (std::string colour)
      : Gdk::Color (Glib::ustring (colour)) {}
    Colour (const Colour& colour)
      : Gdk::Color (colour) {}
    static const Colour& getStartColour () {
      _currentColour.set (ustring ("black"));
      return _currentColour;
    }
    static const Colour& getRandomColour ();
    static const Colour& getNextColour ();
    void setToStartColour () {
      *this = getStartColour ();
    }
    void setToNextColour () {
      *this = getNextColour ();
    }
};

class GC : public Gdk::GC {
  public:
    GC () {}
};


class UI;
class Driver;


class ButtonArray : public Gtk::HBox {
  private:
    std::vector<Colour> _lineColours;
    std::list<PortButton *> _readButtons;
    std::list<PortButton *> _writeButtons;
    std::list<Subscription> _subscriptions;
    Gtk::VBox _readBox;
    Gtk::VBox _writeBox;
    Gtk::DrawingArea _drawingArea;
    
    APB::Driver * _driver;
    APB::UI * _ui;
    
    PortButton * _selectedButton;
    
    void setPorts (Gtk::VBox&, std::list<PortButton *>&, const std::list<Addr *>&, bool);
    bool redrawSubscriptions (GdkEventExpose * event) { return redrawSubscriptions (); }
    bool redrawSubscriptions ();
    const PortButton * findButton (const std::list<PortButton *>&, const Addr *);
    const PortButton * findReadButton (const Addr * addr) {
      return findButton (_readButtons, addr);
    }
    const PortButton * findWriteButton (const Addr * addr) {
      return findButton (_writeButtons, addr);
    }
    void readButtonClicked (PortButton *);
    void writeButtonClicked (PortButton *);
  public:
    ButtonArray (APB::Driver *, APB::UI *);
    Colour& getColour (int);
    void setReadPorts (const std::list<Addr *>&);
    void setWritePorts (const std::list<Addr *>&);
    void setSubscriptions (const std::list<Subscription>&);
    void refreshArray ();
  friend class APB::GTKmm::PortButton;
};

class MainWindow : public Gtk::Window {
  protected:
    LogBox _logBox;
    ButtonArray _buttonArray;
    Gtk::CheckButton _jfdButton;
    APB::Driver * _driver;
  public:
    MainWindow (std::string, APB::Driver *, APB::UI *);
    void log (DebugLevel debugLevel, const std::string& message) {
      _logBox.log (debugLevel, message);
    }
    void jfdToggled ();
    void run () {}
};

#include "apb.h"

class UI : public APB::UI, public SigC::Object {
  Gtk::Main * _gtk;
  MainWindow _mainWindow;
  APB::Driver * _driver;
  std::string _title;
  public:
    UI (std::string, int, char **, APB::Driver *);
    virtual void log (DebugLevel debugLevel, const std::string& message) {
      _mainWindow.log (debugLevel, message);
    }
    virtual void run () {
      _gtk->run(_mainWindow);
    }
    virtual std::vector<ChoiceResult> getChoice (std::string title, const ChoiceSpec& choiceSpec) {
//      ChoiceDialog choiceDialog ("Alsa Patch Bay: Choice", "Subscription flags:",
//                                 choices, _mainWindow);
//      return choiceDialog.getChoice ();
      ChoiceWindow choiceWindow (_title + ": " + title, choiceSpec);
      return choiceWindow.getChoice ();
    }
    void stop () {
      _gtk->quit ();
    }
};

} /* namespace GTKmm */
} /* namespace APB */

#endif /* __APB__GTKMM_H__ */

